# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from dataclasses import dataclass, field, MISSING
from typing import Any, Dict, Type as typingType, Union

from .translate import Translator
from .types import Type, Array, Default, Period

__all__ = [
    # this module
    "Schema",
    # types
    "Type",
    "Array",
    "Default",
    "Period",
    # translate
    "Translator",
]


TypeType = Union[typingType, Type]


@dataclass
class Schema:
    fields: Dict[str, TypeType] = field(default_factory=dict)
    default_fields: Dict[str, Any] = field(init=False)

    def __post_init__(self):
        self.default_fields = {}

        for name, ty in list(self.fields.items()):
            if isinstance(ty, Default):
                self.default_fields[name] = self.fields.pop(name).value

    def apply(self, entry) -> dict:
        result = {
            field_name: ty(entry[field_name])
            for field_name, ty in self.fields.items()
        }

        for name, default in self.default_fields.items():
            result[name] = default
        return result

    def add(self, name: str, ty: TypeType):
        self.fields[name] = ty

        return self

    def add_if(
        self, condition: bool, name: str, ty: TypeType, default=MISSING
    ):
        if not condition:
            if default is not MISSING:
                self.default_fields[name] = ty(default)
            return self

        return self.add(name, ty)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import typing
from dataclasses import dataclass

import pandas as pd
import numpy as np


T = typing.TypeVar("T")


class Type:
    def __call__(self, data):
        raise NotImplementedError


class GenericType(Type, typing.Generic[T]):
    pass


@dataclass
class Default(GenericType[T]):
    value: T
    base: typing.Optional[Type] = None

    def __post_init__(self):
        if self.base is not None:
            self.value = self.base(self.value)

    def __call__(self, data) -> T:
        return self.value


@dataclass
class Array(GenericType[T]):
    """Array type with fixed number of dimensions, but optional dtype and time
    dimension.

    This class ensures that the handled output data, will have `ndim` number of
    dimensions. If specified, `dtype` will be applied to the input to force a
    consistent type, e.g. ``np.float32``. `time_dim` is just a marker,
    indicating which axis notes the time-axis, useful for splitting. If
    `time_dim` is none, the array is time invariant.
    """

    ndim: int
    dtype: typing.Optional[typing.Type[T]] = None
    time_dim: typing.Optional[int] = None

    def __call__(self, data):
        arr = np.asarray(data, dtype=self.dtype)

        if arr.ndim != self.ndim:
            raise ValueError("Dimensions do not match.")

        return arr


@dataclass
class Period:
    freq: typing.Optional[str] = None

    def __call__(self, data):
        return pd.Period(data, freq=self.freq)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import numpy as np
import pandas as pd

from gluonts.dataset.schema import types as ty

f32_2d = ty.Array(dtype=np.float32, ndim=2)


def test_array():
    assert np.array_equal(
        f32_2d([[0, 1, 2]]), np.arange(3, dtype=np.float32).reshape(1, 3)
    )


def test_period():
    assert pd.Period("2020", freq="M") == ty.Period(freq="M")("2020")



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import numpy as np

import pytest

from gluonts.ev import Mean, Sum

VALUE_STREAM = [
    [
        np.full((3, 5), np.nan),
        np.full((3, 5), np.nan),
        np.full((3, 5), np.nan),
    ],
    [
        np.array([[0, np.nan], [0, 0]]),
        np.array([[0, 5], [-5, np.nan]]),
    ],
    [
        np.full(shape=(3, 3), fill_value=1),
        np.full(shape=(1, 3), fill_value=4),
    ],
]

SUM_RES_AXIS_NONE = [
    0,
    0,
    21,
]

SUM_RES_AXIS_0 = [
    np.zeros(5),
    np.array([-5, 5]),
    np.array([7, 7, 7]),
]
SUM_RES_AXIS_1 = [
    np.zeros(9),
    np.array([0, 0, 5, -5]),
    np.array([3, 3, 3, 12]),
]


MEAN_RES_AXIS_NONE = [
    np.nan,
    0,
    1.75,
]

MEAN_RES_AXIS_0 = [
    np.full(5, np.nan),
    np.array([-1.25, 2.5]),
    np.array([1.75, 1.75, 1.75]),
]
MEAN_RES_AXIS_1 = [
    np.full(9, np.nan),
    np.array([0, 0, 2.5, -5]),
    np.array([1, 1, 1, 4]),
]


@pytest.mark.parametrize(
    "value_stream, res_axis_none, res_axis_0, res_axis_1",
    zip(VALUE_STREAM, SUM_RES_AXIS_NONE, SUM_RES_AXIS_0, SUM_RES_AXIS_1),
)
def test_Sum(value_stream, res_axis_none, res_axis_0, res_axis_1):
    for axis, expected_result in zip(
        [None, 0, 1], [res_axis_none, res_axis_0, res_axis_1]
    ):
        sum = Sum(axis=axis)
        for values in value_stream:
            sum.step(np.ma.masked_invalid(values))

        np.testing.assert_almost_equal(sum.get(), expected_result)


@pytest.mark.parametrize(
    "value_stream, res_axis_none, res_axis_0, res_axis_1",
    zip(VALUE_STREAM, MEAN_RES_AXIS_NONE, MEAN_RES_AXIS_0, MEAN_RES_AXIS_1),
)
def test_Mean(value_stream, res_axis_none, res_axis_0, res_axis_1):
    for axis, expected_result in zip(
        [None, 0, 1], [res_axis_none, res_axis_0, res_axis_1]
    ):
        mean = Mean(axis=axis)
        for values in value_stream:
            mean.step(np.ma.masked_invalid(values))

        np.testing.assert_almost_equal(mean.get(), expected_result)


def test_high_dim():
    shape = (4, 9, 5, 2)
    batch_count = 3

    value_stream = [np.random.random(shape) for _ in range(batch_count)]
    all_values = np.concatenate(value_stream)

    for axis in [None, 0, 1, 2, 3]:
        sum = Sum(axis=axis)
        for values in value_stream:
            sum.step(values)

        actual_sum = sum.get()
        expected_sum = all_values.sum(axis=axis)
        np.testing.assert_almost_equal(actual_sum, expected_sum)

        mean = Mean(axis=axis)
        for values in value_stream:
            mean.step(values)

        actual_mean = mean.get()
        expected_mean = all_values.mean(axis=axis)
        np.testing.assert_almost_equal(actual_mean, expected_mean)



import json
import re
import sys
import time
import os
from itertools import chain
from pathlib import Path

import black
import click
import jinja2
import nbformat
import notedown
from nbclient import NotebookClient

from jinja2 import Environment

env = Environment()


def check_github_event(default):
    if "GITHUB_EVENT_PATH" not in os.environ:
        return default

    with open(os.environ["GITHUB_EVENT_PATH"]) as infile:
        event = json.load(infile)

    if "pull_request" in event:
        for label in event["pull_request"]["labels"]:
            if label["name"] == "pr:docs-build-notebook":
                return default

        return "skip"

    return default


def run_notebook(text, kernel_name, timeout) -> str:
    # We add two blank lines at the end to ensure
    # that the final cell also runs.
    text += "\n" * 2
    notebook = notedown.MarkdownReader().reads(text)

    kwargs = {}
    if kernel_name is not None:
        kwargs["kernel_name"] = kernel_name

    client = NotebookClient(
        notebook,
        timeout=timeout,
        resources={"metadata": {"path": "."}},
        **kwargs,
    )
    client.execute()

    # need to add language info to for syntax highlight
    notebook["metadata"].update(language_info={"name": "python"})

    return nbformat.writes(notebook)


def black_cells(text):
    CODE_RE = r"```py(?:thon)?\s*\n(.*?)```"

    text = re.sub(r"^%", r"#%#", text, flags=re.M)

    def apply_black(match):
        code = match.group(1)

        formatted = black.format_str(code, mode=black.Mode())

        return "\n".join(["```", formatted.rstrip(), "```"])

    formatted = re.sub(CODE_RE, apply_black, text, flags=re.S)
    return re.sub(r"^#%#", r"%", formatted, flags=re.M)


def convert(path, mode, kernel_name=None, timeout=40 * 60):
    print(f"=== {path.name} ", end="")
    sys.stdout.flush()

    with path.open() as in_file:
        template = env.from_string(in_file.read())

    markdown = template.render(mode=mode)
    markdown = black_cells(markdown)

    if mode != "skip":
        suffix = ".ipynb"
        start = time.time()
        output = run_notebook(markdown, kernel_name, timeout)
        print(f"finished evaluation in {time.time() - start} sec")
    else:
        suffix = ".md"
        print(f"convert to {suffix}")
        output = markdown

    # XXX.md.input -> XXX.ipynb
    # `with_suffix` only operates on last suffix, so we need some more involved
    # logic.
    stem = path.name.split(".", 1)[0]
    with path.with_name(stem).with_suffix(suffix).open("w") as outfile:
        outfile.write(output)


@click.command()
@click.argument("paths", type=click.Path(), nargs=-1)
@click.option("--kernel", "-k", help="Name of iPython kernel to use.")
@click.option("--mode", "-m", default="release")
def cli(paths, kernel, mode):
    mode = check_github_event(mode)

    for file in map(Path, paths):
        convert(file, kernel_name=kernel, mode=mode)


if __name__ == "__main__":
    cli()
